/**
 * Copyright (c) 2022
 *    Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 *    Abrar (https://gitlab.com/s96Abrar)
 *    rahmanshaber (https://gitlab.com/rahmanshaber)
 *
 * The DFL::Settings expands the QSettings class.
 *
 * IMPL Class
 *  This class will read the various settings files
 *  The class structure and working is inspired by QSettings.
 *  This is purely an implementational detail.
 *  This class can be modified/removed at anytime.
 **/

#pragma once

#include <QMap>
#include <QList>
#include <QString>
#include <QVariant>
#include <QSettings>

#include "DFSettings.hpp"

typedef struct changedsetting_t {
    /** Key whose value was changed */
    QString  key;

    /** Value corresponding to the key */
    QVariant value;

    /** What was the change: Addition (a), Deletion (d), Modification (m) */
    char     state;
} ChangedSetting;

typedef QList<ChangedSetting> ChangedSettings;

namespace DFL {
    namespace Impl {
        class Settings;
        typedef struct setting_backend_t {
            QString        name;
            QString        ext;
            DFL::ReadFunc  rFn;
            DFL::WriteFunc wFn;
        } SettingsBackend;

        typedef QMap<QString, SettingsBackend> SettingsBackends;
    }
}

class DFL::Impl::Settings : public QObject {
    Q_OBJECT;

    public:

        /**
         * @project -> Project like DesQ, Paper, Plasma, etc..
         * @app -> Settings of a specific app: Like Files, Term, Shell, etc...
         * @name -> Name of the settings file. DO NOT include an extension.
         *
         * When using non-scoped settings (i.e., using full path), set
         * project and app as empty strings.
         *
         * No extension has been set at this stage.
         * No config paths have been set at this stage.
         */
        Settings( QString project, QString app, QString name );

        /**
         * Map the dynamic group @dynGrp to an existing group @group
         * See  DFL::Settings::mapDynamicGroup(...) for more details.
         */
        void mapDynamicGroup( QString dynGrp, QString group );

        /**
         * Unmap the previously mapped dynamic group.
         * See  DFL::Settings::unmapDynamicGroup(...) for more details.
         */
        void unmapDynamicGroup( QString dynGrp );

        /**
         * Map a dynamic key to a predefined one.
         * See  DFL::Settings::mapDynamicKey(...) for more details.
         */
        void mapDynamicKey( QString dynKey, QString key );

        /**
         * Unmap the previously mapped dynamic key.
         * See  DFL::Settings::unmapDynamicKey(...) for more details.
         */
        void unmapDynamicKey( QString dynKey );

        /** To get and set the values for a key */
        QVariant value( const QString& key );
        bool setValue( const QString& key, QVariant value );

        /** Keys */
        QStringList allKeys();
        bool hasKey( QString );

        /**
         * Set the settings backend
         */
        bool initializeBackend( SettingsBackend backend );

        /**
         * Setup the config paths
         */
        bool setConfigPath( DFL::Settings::ConfigScope scope, QString cfgPath );

    private:
        /** The watcher to watch for file modifications */
        QFileSystemWatcher *watcher;

        /** The selected backend */
        SettingsBackend *backend = nullptr;

        /** Project, Application and Settings file names */
        QString mProjectName;
        QString mAppName;
        QString mSettingFileName;

        /** Config paths */
        QString defCfgPath;     // Default: /usr/share/<org>/<app>/<file>.conf
        QString sysCfgPath;     // Default: /etc/xdg/<org>/<app>/<file>.conf
        QString usrCfgPath;     // Default: ~/.config/<org>/<app>/<file>.conf

        /** Settings Maps */
        QVariantMap defSett;    // Developer defaults
        QVariantMap sysSett;    // Distro defaults
        QVariantMap usrSett;    // User settings

        /** This is the final settings map of the app */
        QVariantMap map;

        /** Container to store all the registered backends */
        QMap<int, QString> mConfigPaths;

        /** Dynamic groups and keys */
        QMap<QString, QString> mDynamicGroups;
        QMap<QString, QString> mDynamicKeys;

        /** Flag to mark the completion of initializations */
        bool mInitComplete = false;

        /** Initialize all the components */
        void initialize();

        /** Load the current settings. Responsible for maintaining usr/sys/def hierarchy */
        QVariantMap readCurrentSettings();

        /** Create the default settings backend */
        void initializeDefaultBackend();

        /**
         * Get the changed settings as a list
         * Read the current settings, then compare with the settings stored in @map.
         * Store the changed settings, and update @map.
         * Finally return the changed settings.
         */
        void changedSettings( QString file );

    Q_SIGNALS:
        void settingsChanged( ChangedSettings );
};
